function [flights,M,minima] = UTM_zipper(max_steps,H_X,d_S,max_time)
% UTM_zipper - real-time deconfliction protocol simulation
% On input:
%     max_steps (int): number of steps in simulation
%     H_X (float): minimum lane headway distance
%     d_S (float): minimum Euclidean headway distance
%     max_time (float): max time for sim clock
% On output:
%     flights (flight struct): flight info
%       .status (int): 1 (WAITING); 0 (IN_FLIGHT); -1 (DONE)
%       .lane_index (int): index into current lane sequence
%       .lane_dists (1xn vector): distance travelled along each lane
%       .lane_speeds (1xn vector): final speed in each lane
%       .start_time (float): actua launch time
%       .original_start_time (float): desired launch time
%       .launch_node (int): index of launch lane (node?)
%       .land_node (int): index of land lane (node?)
%       .traj (1xn vector): lane indexes of trajectory
%       .lane_lengths(1xn vector): lengths of lanes in trajectory
%       .TOA (1xn vector): Time of Arrival at exit point of lane
%       .TOD (1xn vector): Time of Departure from entry point of each lane
%       .last_update (float): time of last state update for flight
%       .total_dist (float): total distance tavelled to current time
%       .num_restarts (int): number of time launch postponed
%     M (Matlab movie): movie of flights through lanes
% Call:
%     [res,M] = UTM_zipper(500,1.4142,1,10);
% Author:
%     T. Henderson
%     UU
%     Summer 2020
%

UTM_load_vars;

flights = [];
minima = [];

% Set up lanes:
%   nodes (kx3 array): locations of lane endpoints
%   edges (mx2 array): lanes defined as directed edges between nodes
%   paths(kxk struct array).traj: sequence of node indexes from launch
%     nodes to land nodes
%   trajectories(kxk struct array): lane index sequence between nodes
%   lanes (nx6 array): lane entry point, exit point per row
%   co_lanes (struct vector).lanes: lane indexes connected to each lane (?)
%   launch_nodes (1xp vector): launch node/lane indexes
%   land_nodes (1xq vector): land node/lane indexes
[nodes,edges,paths,trajectories,lanes,lane_degrees,angles,co_lanes,...
    launch_nodes,land_nodes,lane_paths] = UTM_zipper_lanes;
num_nodes = length(nodes(:,1));
num_edges = length(edges(:,1));
num_lanes = length(lanes(:,1));
num_launch_nodes = length(launch_nodes);
num_land_nodes = length(land_nodes);
lane_lengths = zeros(num_lanes,1);
for k = 1:num_lanes
    lane_lengths(k) = norm(lanes(k,1:3)-lanes(k,4:6));
end

num_flights = 0;
m_count = 0;
cur_time = 0;
s = 0;

while cur_time<max_time
    s = s + 1;
    [cur_time];
    if rand<NEW_FLIGHT_PROB    % generate a flight
        num_flights = num_flights + 1;
        f = num_flights;
        start_time = cur_time + MIN_DELAY + rand*(MAX_DELAY-MIN_DELAY);
%        start_time = s*0.5;
        node1 = launch_nodes(randi(num_launch_nodes));
        node2 = land_nodes(randi(num_land_nodes));
%        node1 = 1;
%        node2 = 20;
        traj = trajectories(node1,node2).traj;
        traj = lane_paths(node1,node2).lanes;
        num_lanes = length(traj);
        speed = MIN_SPEED + rand*(MAX_SPEED-MIN_SPEED);
        speed = MAX_SPEED; % always set maximum speed
        flights(f).status = WAITING;
        flights(f).lane_index = 1;
        flights(f).lane_dists = zeros(1,num_lanes);
        lane_speeds = speed*ones(1,num_lanes);
        flights(f).lane_speeds = lane_speeds;
        flights(f).start_time = start_time;
        flights(f).original_start_time = start_time;
        flights(f).launch_node = node1;
        flights(f).land_node = node2;
        flights(f).traj = traj;
        flights(f).lane_lengths = lane_lengths(traj);
        [TOD,TOA] = UTM_set_TOD_TOA(start_time,lane_speeds,lane_lengths);
        flights(f).TOD = TOD;
        flights(f).TOA = TOA;
        flights(f).last_update = cur_time;
        flights(f).total_dist = 0;
        flights(f).num_restarts = 0;
    end
    prev_time = cur_time;
    [cur_time,f] = UTM_next_event(flights);
    if f>0 % handle next event
        
        %advance flights
        flights = UTM_advance_flights(flights,lanes,cur_time);

        % handle selected event
        lane_index = flights(f).lane_index;
        lane_number = flights(f).traj(lane_index);
        if flights(f).status==WAITING
            [min_dist,min_f] = UTM_last_in_lane(flights,lane_number);
            flights(f).last_update = cur_time;
            if min_dist<H_X
                % reschedule: too close
                start_time = cur_time + MIN_DELAY...
                    + rand*(MAX_DELAY-MIN_DELAY);
                flights(f).start_time = start_time;
                num_lanes = length(flights(f).traj);
                [TOD,TOA] = UTM_set_TOD_TOA(start_time,...
                    flights(f).lane_speeds,flights(f).lane_lengths);
                flights(f).TOD = TOD;
                flights(f).TOA = TOA;
                flights(f).num_restarts = flights(f).num_restarts + 1;
            else
                flights(f).status = IN_FLIGHT;
                [flights,slowed] = UTM_Ben_deconflict(flights,lanes,...
                    co_lanes,H_X,cur_time);
%                flights = UTM_zipper_TD(flights,lanes,lane_degrees,...
%                    co_lanes,edges,f,H_X,cur_time,...
%                    flights(f).lane_dists(lane_index));
            end
        else % deconflict at new lane
            CLOSED = [];
            queue = f;
            while ~isempty(queue)
                fp = queue(1);
                queue = queue(2:end);
                fp_lane_index = flights(fp).lane_index;
                fp_lane = flights(fp).traj(fp_lane_index);
                CLOSED = [CLOSED,fp];
%                old_speeds = flights(f).lane_speeds;
                [flights,slowed] = UTM_Ben_deconflict(flights,lanes,...
                    co_lanes,H_X,cur_time);
%                flights = UTM_zipper_TD(flights,lanes,lane_degrees,co_lanes,...
%                    edges,f,H_X,cur_time,flights(f).lane_dists(lane_index));
%                new_speeds = flights(f).lane_speeds;
%                slowed = abs(new_speeds(lane_index)-old_speeds(lane_index))...
%                    >ZERO_THRESH;
%                 if slowed==1
%                     c_lanes = [co_lanes(fp_lane).entry_in,...
%                         co_lanes(fp_lane).entry_out,...
%                         co_lanes(fp_lane).exit_in];
%                     num_c_lanes = length(c_lanes);
%                     c_flights = [];
%                     for k = 1:num_c_lanes
%                         f_info = UTM_flights_in_lane(flights,c_lanes(k));
%                         if ~isempty(f_info)
%                             num_f_info = length(f_info(:,1));
%                             for n = 1:num_f_info
%                                 c_flights = [c_flights,f_info(n,1)];
%                             end
%                         end
%                     end
% %                    c_flights = setdiff(c_flights,CLOSED);
%                     c_flights = setdiff(c_flights,queue);
%                     queue = [queue,c_flights];
%                 end
            end
        end
%         for f1 = 1:num_flights-1
%             f1i = flights(f1).lane_index;
%             for f2 = f1+1:num_flights
%                 f2i = flights(f2).lane_index;
%                 min_d = 100;
%                 min_t = 100;
%                 if flights(f1).status==0&flights(f2).status==0
%                     [min_d,min_t] = UTM_2conflict(flights,lanes,f1,f2,...
%                         f1i,f2i,cur_time);
%                     if min_d<d_S/2&min_t<max_time
%                         if f1i<f2i
%                             flights(f1).status = DONE;
%                         elseif f2i<=f1i
%                             flights(f2).status = DONE;
%                         end
%                     end
%                 end
%                 tch = 0;
%             end
%         end
        UTM_draw_zipper(flights,lanes);
        text(0,0,-1,num2str(s));
        drawnow
        m_count = m_count + 1;
        M(m_count) = getframe(gca);
        dists = UTM_zipper_dists(flights,lanes);
        min_val = UTM_min_dist(dists);
        minima = [minima,min_val];
        [cur_time min_val]
        cc = UTM_conflicts(flights,lanes);
        if min_val+ZERO_THRESH<d_S/2
            display(['Flights too close: Step: ',num2str(s)]);
        end
    end
end
res = flights;

tch = 0;
